@@ -10,26 +10,31 @@
     January 22, 2020
 """
 ################# GLOBAL IMPORTS ####################
+import warnings
+from tabulate import tabulate
+import shutil
+from soxspipe.commonutils import filenamer
+from datetime import datetime
+from soxspipe.commonutils import detector_lookup
+from soxspipe.commonutils import keyword_lookup
+from soxspipe.commonutils import set_of_files
+from ccdproc import Combiner
+from astropy.nddata.nduncertainty import StdDevUncertainty
+import pandas as pd
+from soxspipe.commonutils import subtract_background
+import ccdproc
+from astropy.stats import sigma_clip, mad_std
+from astropy import units as u
+from astropy.nddata import CCDData
+import numpy as np
+from fundamentals import tools
 from builtins import object
 import sys
+import math
+import inspect
+import yaml
 import os
 os.environ['TERM'] = 'vt100'
-from fundamentals import tools
-import numpy as np
-from astropy.nddata import CCDData
-from astropy import units as u
-from astropy.stats import mad_std
-import ccdproc
-import pandas as pd
-from astropy.nddata.nduncertainty import StdDevUncertainty
-from ccdproc import Combiner
-from soxspipe.commonutils import set_of_files
-from soxspipe.commonutils import keyword_lookup
-from soxspipe.commonutils import detector_lookup
-from datetime import datetime
-from soxspipe.commonutils import filenamer
-import shutil
-from tabulate import tabulate
 
 
 class _base_recipe_(object):
@@ -59,8 +64,8 @@ class _base_recipe_(object):
             settings["intermediate-data-root"])
         self.reducedRootPath = self._absolute_path(
             settings["reduced-data-root"])
-        self.calibrationRootPath = self._absolute_path(
-            settings["calibration-data-root"])
+        from soxspipe.commonutils.toolkit import get_calibrations_path
+        self.calibrationRootPath = get_calibrations_path(log=self.log, settings=self.settings)
 
         self.verbose = verbose
         # SET LATER WHEN VERIFYING FRAMES
@@ -68,14 +73,37 @@ class _base_recipe_(object):
         self.detectorParams = None
         self.dateObs = None
 
+        # COLLECT ADVANCED SETTINGS IF AVAILABLE
+        parentDirectory = os.path.dirname(__file__)
+        advs = parentDirectory + "/advanced_settings.yaml"
+        level = 0
+        exists = False
+        count = 1
+        while not exists and len(advs) and count < 10:
+            count += 1
+            level -= 1
+            exists = os.path.exists(advs)
+            if not exists:
+                advs = "/".join(parentDirectory.split("/")
+                                [:level]) + "/advanced_settings.yaml"
+        if not exists:
+            advs = {}
+        else:
+            with open(advs, 'r') as stream:
+                advs = yaml.safe_load(stream)
+        # MERGE ADVANCED SETTINGS AND USER SETTINGS (USER SETTINGS OVERRIDE)
+        self.settings = {**advs, **self.settings}
+
         # DATAFRAMES TO COLLECT QCs AND PRODUCTS
         self.qc = pd.DataFrame({
             "soxspipe_recipe": [],
             "qc_name": [],
             "qc_value": [],
             "qc_unit": [],
+            "qc_comment": [],
             "obs_date_utc": [],
-            "reduction_date_utc": []
+            "reduction_date_utc": [],
+            "to_header": []
         })
         self.products = pd.DataFrame({
             "soxspipe_recipe": [],
@@ -116,6 +144,9 @@ class _base_recipe_(object):
         """
         self.log.debug('starting the ``_prepare_single_frame`` method')
 
+        warnings.filterwarnings(
+            'ignore')
+
         kw = self.kw
         dp = self.detectorParams
 
@@ -127,8 +158,12 @@ class _base_recipe_(object):
             # CONVERT RELATIVE TO ABSOLUTE PATHS
             frame = self._absolute_path(frame)
             # OPEN THE RAW FRAME - MASK AND UNCERT TO BE POPULATED LATER
-            frame = CCDData.read(frame, hdu=0, unit=u.adu, hdu_uncertainty='ERRS',
-                                 hdu_mask='QUAL', hdu_flags='FLAGS', key_uncertainty_type='UTYPE')
+            try:
+                frame = CCDData.read(frame, hdu=0, unit=u.adu, hdu_uncertainty='ERRS',
+                                     hdu_mask='QUAL', hdu_flags='FLAGS', key_uncertainty_type='UTYPE')
+            except TypeError as e:
+                self.log.info(f"{filepath} is a FITS Binary Table")
+                return filepath
 
         # CHECK THE NUMBER OF EXTENSIONS IS ONLY 1 AND "SXSPRE" DOES NOT
         # EXIST. i.e. THIS IS A RAW UNTOUCHED FRAME
@@ -139,6 +174,19 @@ class _base_recipe_(object):
         frame = self.xsh2soxs(frame)
         frame = self._trim_frame(frame)
 
+        # FUDGE BAD-PIXEL MAP CREATION
+        # bpmData = np.random.rand(frame.data.shape[0], frame.data.shape[1])
+        # bpmData[bpmData < 0.995] = 0
+        # bpmData[bpmData > 0] = 1
+        # bpmData = CCDData(bpmData, unit="adu")
+        # from soxspipe.commonutils.toolkit import quicklook_image
+        # quicklook_image(
+        #     log=self.log, CCDObject=bpmData, show=False, ext='data', stdWindow=3)
+        # HDUList = bpmData.to_hdu()
+        # HDUList[0].name = "BPM"
+        # HDUList.writeto("/Users/Dave/Desktop/tmp.fits", output_verify='exception',
+        #                 overwrite=True, checksum=True)
+
         # CORRECT FOR GAIN - CONVERT DATA FROM ADU TO ELECTRONS
         frame = ccdproc.gain_correct(frame, dp["gain"])
 
@@ -157,7 +205,7 @@ class _base_recipe_(object):
         # NOTE FLAGS NOTE YET SUPPORTED BY CCDPROC THIS THIS WON'T GET SAVED OUT
         # AS AN EXTENSION
         arm = self.arm
-        if kw('WIN_BINX') in frame.header:
+        if arm != "NIR" and kw('WIN_BINX') in frame.header:
             binx = int(frame.header[kw('WIN_BINX')])
             biny = int(frame.header[kw('WIN_BINY')])
         else:
@@ -175,8 +223,8 @@ class _base_recipe_(object):
 
         # BIAS FRAMES HAVE NO 'FLUX', JUST READNOISE, SO ADD AN EMPTY BAD-PIXEL
         # MAP
-        if frame.header[kw("DPR_TYPE")] == "BIAS":
-            bitMap.data = np.zeros_like(bitMap.data)
+        # if frame.header[kw("DPR_TYPE")] == "BIAS":
+        #     bitMap.data = np.zeros_like(bitMap.data)
 
         # print(bitMap.data.shape)
         # print(frame.data.shape)
@@ -284,10 +332,12 @@ class _base_recipe_(object):
         filepaths = self.inputFrames.files_filtered(include_path=True)
 
         frameCount = len(filepaths)
+
         print("\n# PREPARING %(frameCount)s RAW FRAMES - TRIMMING OVERSCAN, CONVERTING TO ELECTRON COUNTS, GENERATING UNCERTAINTY MAPS AND APPENDING DEFAULT BAD-PIXEL MASK" % locals())
         preframes = []
         preframes[:] = [self._prepare_single_frame(
             frame=frame, save=save) for frame in filepaths]
+
         sof = set_of_files(
             log=self.log,
             settings=self.settings,
@@ -325,7 +375,7 @@ class _base_recipe_(object):
             keyword=kw("SEQ_ARM").lower(), unique=True)
         # MIXED INPUT ARMS ARE BAD
         if len(arm) > 1:
-            arms = " and ".join(arms)
+            arms = " and ".join(arm)
             print(self.inputFrames.summary)
             raise TypeError(
                 "Input frames are a mix of %(imageTypes)s" % locals())
@@ -340,13 +390,22 @@ class _base_recipe_(object):
         ).get(self.arm)
 
         # MIXED BINNING IS BAD
-        cdelt1 = self.inputFrames.values(
-            keyword=kw("CDELT1").lower(), unique=True)
-        cdelt2 = self.inputFrames.values(
-            keyword=kw("CDELT2").lower(), unique=True)
+        if self.arm == "NIR":
+            # NIR ARRAY NEVER BINNED
+            cdelt1 = [1]
+            cdelt2 = [1]
+        else:
+            cdelt1 = self.inputFrames.values(
+                keyword=kw("CDELT1").lower(), unique=True)
+            cdelt2 = self.inputFrames.values(
+                keyword=kw("CDELT2").lower(), unique=True)
+            try:
+                cdelt1.remove(None)
+                cdelt2.remove(None)
+            except:
+                pass
 
         if len(cdelt1) > 1 or len(cdelt2) > 1:
-            print(self.inputFrames.summary)
             raise TypeError(
                 "Input frames are a mix of binnings" % locals())
 
@@ -356,21 +415,32 @@ class _base_recipe_(object):
         # MIXED READOUT SPEEDS IS BAD
         readSpeed = self.inputFrames.values(
             keyword=kw("DET_READ_SPEED").lower(), unique=True)
+        from contextlib import suppress
+        with suppress(ValueError):
+            readSpeed.remove(None)
+
         if len(readSpeed) > 1:
             print(self.inputFrames.summary)
             raise TypeError(
-                "Input frames are a mix of readout speeds" % locals())
+                f"Input frames are a mix of readout speeds. {readSpeed}" % locals())
 
         # MIXED GAIN SPEEDS IS BAD
         # HIERARCH ESO DET OUT1 CONAD - Electrons/ADU
         # CONAD IS REALLY GAIN AND HAS UNIT OF Electrons/ADU
-        gain = self.inputFrames.values(
-            keyword=kw("CONAD").lower(), unique=True)
+        if self.settings["instrument"] == "xsh":
+            gain = self.inputFrames.values(
+                keyword=kw("CONAD").lower(), unique=True)
+        else:
+            gain = self.inputFrames.values(
+                keyword=kw("GAIN").lower(), unique=True)
+
+        with suppress(ValueError):
+            gain.remove(None)
         if len(gain) > 1:
             print(self.inputFrames.summary)
             raise TypeError(
                 "Input frames are a mix of gain" % locals())
-        if gain[0]:
+        if len(gain) and gain[0]:
             # UVB & VIS
             self.detectorParams["gain"] = gain[0] * u.electron / u.adu
         else:
@@ -381,12 +451,14 @@ class _base_recipe_(object):
         # HIERARCH ESO DET OUT1 RON - Readout noise in electrons
         ron = self.inputFrames.values(
             keyword=kw("RON").lower(), unique=True)
+        with suppress(ValueError):
+            ron.remove(None)
 
         # MIXED NOISE
         if len(ron) > 1:
             print(self.inputFrames.summary)
-            raise TypeError("Input frames are a mix of readnoise" % locals())
-        if ron[0]:
+            raise TypeError(f"Input frames are a mix of readnoise. {ron}" % locals())
+        if len(ron) and ron[0]:
             # UVB & VIS
             self.detectorParams["ron"] = ron[0] * u.electron
         else:
@@ -394,8 +466,35 @@ class _base_recipe_(object):
             self.detectorParams["ron"] = self.detectorParams[
                 "ron"] * u.electron
 
+        imageTypes = self.inputFrames.values(
+            keyword=kw("DPR_TYPE").lower(), unique=True) + self.inputFrames.values(
+            keyword=kw("PRO_TYPE").lower(), unique=True)
+        imageTech = self.inputFrames.values(
+            keyword=kw("DPR_TECH").lower(), unique=True) + self.inputFrames.values(
+            keyword=kw("PRO_TECH").lower(), unique=True)
+        imageCat = self.inputFrames.values(
+            keyword=kw("DPR_CATG").lower(), unique=True) + self.inputFrames.values(
+            keyword=kw("PRO_CATG").lower(), unique=True)
+
+        def clean_list(myList):
+            myList = list(set(myList))
+            try:
+                myList.remove(None)
+            except:
+                pass
+            try:
+                myList.remove("REDUCED")
+            except:
+                pass
+
+            return myList
+
+        imageTypes = clean_list(imageTypes)
+        imageTech = clean_list(imageTech)
+        imageCat = clean_list(imageCat)
+
         self.log.debug('completed the ``_verify_input_frames_basics`` method')
-        return None
+        return imageTypes, imageTech, imageCat
 
     def clean_up(
             self):
@@ -504,6 +603,25 @@ class _base_recipe_(object):
         """
         self.log.debug('starting the ``write`` method')
 
+        # WRITE QCs TO HEADERS
+        for n, v, c, h in zip(self.qc["qc_name"].values, self.qc["qc_value"].values, self.qc["qc_comment"].values, self.qc["to_header"].values):
+            if h:
+                frame.header[f"ESO QC {n}".upper()] = (v, c)
+
+        # NEATLY SORT KEYWORDS
+        keywords = [k for k in frame.header if len(k)]
+        values = [frame.header[k] for k in frame.header if len(k)]
+        comments = [frame.header.comments[k] for k in frame.header if len(k)]
+        keywords, values, comments = zip(
+            *sorted(zip(keywords, values, comments)))
+        if "COMMENT" not in keywords and "HISTORY" not in keywords:
+            frame.header.clear()
+            for k, v, c in zip(keywords, values, comments):
+                if k == "COMMENT":
+                    frame.header[k] = v
+                else:
+                    frame.header[k] = (v, c)
+
         if not filename:
 
             filename = filenamer(
@@ -517,7 +635,7 @@ class _base_recipe_(object):
         HDUList = frame.to_hdu(
             hdu_mask='QUAL', hdu_uncertainty='ERRS', hdu_flags=None)
         HDUList[0].name = "FLUX"
-        HDUList.writeto(filepath, output_verify='exception',
+        HDUList.writeto(filepath, output_verify='fix+warn',
                         overwrite=overwrite, checksum=True)
 
         filepath = os.path.abspath(filepath)
@@ -565,6 +683,11 @@ class _base_recipe_(object):
         # ALLOW FOR UNDERSCORE AND HYPHENS
         recipe = recipe.replace("soxs_", "soxs-")
 
+        # UNPACK SETTINGS
+        stacked_clipping_sigma = self.settings[
+            recipe]["stacked-clipping-sigma"]
+        stacked_clipping_iterations = self.settings[
+            recipe]["stacked-clipping-iterations"]
         # UNPACK SETTINGS
         clipping_lower_sigma = self.settings[
             recipe]["clipping-lower-simga"]
@@ -576,7 +699,7 @@ class _base_recipe_(object):
         # LIST OF CCDDATA OBJECTS NEEDED BY COMBINER OBJECT
         if not isinstance(frames, list):
             ccds = [c for c in frames.ccds(ccd_kwargs={"hdu_uncertainty": 'ERRS',
-                                                       "hdu_mask": 'QUAL', "hdu_flags": 'FLAGS', "key_uncertainty_type": 'UTYPE'})]
+                                                       "hdu_mask": 'QUAL', "hdu_flags": 'FLAGS', "key_uncertainty_type": 'UTYPE', "unit": u.electron})]
         else:
             ccds = frames
 
@@ -610,9 +733,9 @@ class _base_recipe_(object):
         # THIS IS THE SUM OF BAD-PIXELS IN ALL INDIVIDUAL FRAME MASKS
         new_n_masked = combiner.data_arr.mask.sum()
         iteration = 1
-        while (new_n_masked > old_n_masked and iteration <= clipping_iteration_count):
+        while (new_n_masked > old_n_masked and iteration <= stacked_clipping_iterations):
             combiner.sigma_clipping(
-                low_thresh=clipping_lower_sigma, high_thresh=clipping_upper_sigma, func=np.ma.median, dev_func=mad_std)
+                low_thresh=stacked_clipping_sigma, high_thresh=stacked_clipping_sigma, func=np.ma.median, dev_func=mad_std)
             old_n_masked = new_n_masked
             # RECOUNT BAD-PIXELS NOW CLIPPING HAS RUN
             new_n_masked = combiner.data_arr.mask.sum()
@@ -631,6 +754,13 @@ class _base_recipe_(object):
         # RECOMBINE THE COMBINED MASK FROM ABOVE
         combined_frame.mask = combined_frame.mask | combinedMask
 
+        # NOW SIMGA-CLIP ACROSS THE FRAME
+        # SIGMA-CLIP THE DATA (AT HIGH LEVEL)
+        if clipping_iteration_count:
+            maskedFrame = sigma_clip(
+                combined_frame, sigma_lower=clipping_lower_sigma, sigma_upper=clipping_upper_sigma, maxiters=clipping_iteration_count, cenfunc='median', stdfunc=mad_std)
+            combined_frame.mask = combined_frame.mask | maskedFrame.mask
+
         # MASSIVE FUDGE - NEED TO CORRECTLY WRITE THE HEADER FOR COMBINED
         # IMAGES
         combined_frame.header = ccds[0].header
@@ -644,32 +774,42 @@ class _base_recipe_(object):
         # CALCULATE NEW PIXELS ADDED TO MASK
         newBadCount = combined_frame.mask.sum()
         diff = newBadCount - badCount
-        print("\t%(diff)s new pixels made it into the combined bad-pixel map" % locals())
+        totalPixels = np.size(combinedMask)
+        percent = (float(newBadCount) / float(totalPixels)) * 100.
+        print(f"\t{diff} new pixels made it into the combined bad-pixel map (bad pixels now account for {percent:0.2f}% of all pixels)")
+
+        # print(combined_frame.uncertainty)
+        # print(combined_frame.data)
+        # print(np.multiply(combined_frame.uncertainty,combined_frame.data))
 
         self.log.debug('completed the ``clip_and_stack`` method')
         return combined_frame
 
-    def subtract_calibrations(
+    def detrend(
             self,
             inputFrame,
             master_bias=False,
-            dark=False):
+            dark=False,
+            master_flat=False,
+            order_table=False):
         """*subtract calibration frames from an input frame*
 
         **Key Arguments:**
             - ``inputFrame`` -- the input frame to have calibrations subtracted. CCDData object.
             - ``master_bias`` -- the master bias frame to be subtracted. CCDData object. Default *False*.
             - ``dark`` -- a dark frame to be subtracted. CCDData object. Default *False*.
+            - ``master_flat`` -- divided input frame by this master flat frame. CCDData object. Default *False*.
+            - ``order_table`` -- order table with order edges defined. Used to subtract scattered light background from frames. Default *False*.
 
         **Return:**
             - ``calibration_subtracted_frame`` -- the input frame with the calibration frame(s) subtracted. CCDData object.
 
         **Usage:**
 
-        Within a soxspipe recipe use `subtract_calibrations` like so:
+        Within a soxspipe recipe use `detrend` like so:
 
         ```python
-        myCalibratedFrame = self.subtract_calibrations(
+        myCalibratedFrame = self.detrend(
             inputFrame=inputFrameCCDObject, master_bias=masterBiasCCDObject, dark=darkCCDObject)
         ```
 
@@ -681,7 +821,7 @@ class _base_recipe_(object):
             - code needs written to scale dark frame to exposure time of science/calibration frame
         ```
         """
-        self.log.debug('starting the ``subtract_calibrations`` method')
+        self.log.debug('starting the ``detrend`` method')
 
         arm = self.arm
         kw = self.kw
@@ -695,7 +835,7 @@ class _base_recipe_(object):
         # VERIFY DATA IS IN ORDER
         if master_bias == False and dark == False:
             raise TypeError(
-                "subtract_calibrations method needs a master-bias frame and/or a dark frame to subtract")
+                "detrend method needs at least a master-bias frame and/or a dark frame to subtract")
         if master_bias == False and dark.header[kw("EXPTIME")] != inputFrame.header[kw("EXPTIME")]:
             raise AttributeError(
                 "Dark and science/calibration frame have differing exposure-times. A master-bias frame needs to be supplied to scale the dark frame to same exposure time as input science/calibration frame")
@@ -703,54 +843,43 @@ class _base_recipe_(object):
             raise AttributeError(
                 "CODE NEEDS WRITTEN HERE TO SCALE DARK FRAME TO EXPOSURE TIME OF SCIENCE/CALIBRATION FRAME")
 
+        processedFrame = inputFrame.copy()
+
+        if master_bias != False:
+            processedFrame = ccdproc.subtract_bias(processedFrame, master_bias)
+
         # DARK WITH MATCHING EXPOSURE TIME
         tolerence = 0.5
-        if dark != False and (int(dark.header[kw("EXPTIME")]) < int(inputFrame.header[kw("EXPTIME")]) + tolerence) and (int(dark.header[kw("EXPTIME")]) > int(inputFrame.header[kw("EXPTIME")]) - tolerence):
-            calibration_subtracted_frame = inputFrame.subtract(dark)
-            calibration_subtracted_frame.header = inputFrame.header
-            try:
-                calibration_subtracted_frame.wcs = inputFrame.wcs
-            except:
-                pass
+        if dark != False and (int(dark.header[kw("EXPTIME")]) < int(processedFrame.header[kw("EXPTIME")]) + tolerence) and (int(dark.header[kw("EXPTIME")]) > int(processedFrame.header[kw("EXPTIME")]) - tolerence):
+            processedFrame = ccdproc.subtract_bias(processedFrame, dark)
 
-        # ONLY A MASTER BIAS FRAME, NO DARK
-        if dark == False and master_bias != False:
-            calibration_subtracted_frame = inputFrame.subtract(master_bias)
-            calibration_subtracted_frame.header = inputFrame.header
-            try:
-                calibration_subtracted_frame.wcs = inputFrame.wcs
-            except:
-                pass
+        if master_flat != False:
+            processedFrame = ccdproc.flat_correct(processedFrame, master_flat)
 
-        self.log.debug('completed the ``subtract_calibrations`` method')
-        return calibration_subtracted_frame
+        if order_table != False and 1 == 1:
+            background = subtract_background(
+                log=self.log,
+                frame=processedFrame,
+                orderTable=order_table,
+                settings=self.settings
+            )
+            backgroundFrame, processedFrame = background.subtract()
+
+        self.log.debug('completed the ``detrend`` method')
+        return processedFrame
 
     def report_output(
             self,
             rformat="stdout"):
-        """*a method to report QC values alongside imtermediate and final products*
+        """*a method to report QC values alongside intermediate and final products*
 
         **Key Arguments:**
             - ``rformat`` -- the format to outout reports as. Default *stdout*. [stdout|....]
 
-        **Return:**
-            - None
-
         **Usage:**
 
         ```python
-        usage code 
-        ```
-
-        ---
-
-        ```eval_rst
-        .. todo::
-
-            - add usage info
-            - create a sublime snippet for usage
-            - write a command-line tool for this method
-            - update package tutorial with command-line tool info if needed
+        self.report_output(rformat="stdout")
         ```
         """
         self.log.debug('starting the ``report_output`` method')
@@ -759,12 +888,158 @@ class _base_recipe_(object):
             # REMOVE COLUMN FROM DATA FRAME
             self.products.drop(columns=['file_path'], inplace=True)
 
+        columns = list(self.qc.columns)
+        columns.remove("to_header")
+
         if rformat == "stdout":
-            print(tabulate(self.qc, headers='keys', tablefmt='psql'))
-            print(tabulate(self.products, headers='keys', tablefmt='psql'))
+            print("\n# QCs")
+            print(tabulate(self.qc[columns], headers='keys', tablefmt='psql', showindex=False, stralign="right"))
+            print("\n# RECIPE PRODUCTS")
+            print(tabulate(self.products, headers='keys', tablefmt='psql', showindex=False, stralign="right"))
 
         self.log.debug('completed the ``report_output`` method')
         return None
 
+    def qc_ron(
+            self,
+            frameType=False,
+            frameName=False,
+            masterFrame=False):
+        """*calculate the read-out-noise from bias/dark frames*
+
+        **Key Arguments:**
+            - ``frameType`` -- the type of the frame for reporting QC values. Default *False*
+            - ``frameName`` -- the name of the frame in human readable words. Default *False*
+            - ``masterFrame`` -- the master frame (only makes sense to measure RON on master bias). Default *False*
+
+        **Return:**
+            - ``rawRon`` -- raw read-out-noise in electrons
+            - ``masterRon`` -- combined read-out-noise in mbias
+
+        **Usage:**
+
+        ```python
+        rawRon, mbiasRon = self.qc_ron(
+            frameType="MBIAS",
+            frameName="master bias",
+            masterFrame=masterFrame
+        )
+        ```
+        """
+        self.log.debug('starting the ``qc_bias_ron`` method')
+
+        utcnow = datetime.utcnow()
+        utcnow = utcnow.strftime("%Y-%m-%dT%H:%M:%S")
+
+        # LIST OF RAW CCDDATA OBJECTS
+        ccds = [c for c in self.inputFrames.ccds(ccd_kwargs={
+                                                 "hdu_uncertainty": 'ERRS', "hdu_mask": 'QUAL', "hdu_flags": 'FLAGS', "key_uncertainty_type": 'UTYPE'})]
+
+        # SINGLE FRAME RON
+        raw_one = ccds[0]
+        raw_two = ccds[1]
+        raw_diff = raw_one.subtract(raw_two)
+
+        # SIGMA-CLIP THE DATA (AT HIGH LEVEL)
+        masked_diff = sigma_clip(
+            raw_diff, sigma_lower=10, sigma_upper=10, maxiters=5, cenfunc='median', stdfunc=mad_std)
+        combinedMask = raw_diff.mask | masked_diff.mask
+
+        # FORCE CONVERSION OF CCDData OBJECT TO NUMPY ARRAY
+        raw_diff = np.ma.array(raw_diff.data, mask=combinedMask)
+
+        def imstats(dat): return (dat.min(), dat.max(), dat.mean(), dat.std())
+        dmin, dmax, dmean, dstd = imstats(raw_diff)
+
+        # ACCOUNT FOR EXTRA NOISE ADDED FROM SUBTRACTING FRAMES
+        rawRon = dstd / math.sqrt(2)
+
+        singleFrameType = frameType
+        if frameType[0] == "M":
+            singleFrameType = frameType[1:]
+
+        self.qc = self.qc.append({
+            "soxspipe_recipe": self.recipeName,
+            "qc_name": "RON DETECTOR",
+            "qc_value": rawRon,
+            "qc_comment": f"RON in single {singleFrameType} (electrons)",
+            "qc_unit": "electrons",
+            "obs_date_utc": self.dateObs,
+            "reduction_date_utc": utcnow,
+            "to_header": True
+        }, ignore_index=True)
+
+        if masterFrame:
+
+            # PREDICTED MASTER NOISE
+            predictedMasterRon = rawRon / math.sqrt(len(ccds))
+
+            dmin, dmax, dmean, dstd = imstats(masterFrame.data)
+            masterRon = dstd
+
+            self.qc = self.qc.append({
+                "soxspipe_recipe": self.recipeName,
+                "qc_name": "RON MASTER",
+                "qc_value": masterRon,
+                "qc_comment": f"Combined RON in {frameType} (electrons)",
+                "qc_unit": "electrons",
+                "obs_date_utc": self.dateObs,
+                "reduction_date_utc": utcnow,
+                "to_header": True
+            }, ignore_index=True)
+        else:
+            masterRon = None
+
+        self.log.debug('completed the ``qc_bias_ron`` method')
+        return rawRon, masterRon
+
+    def qc_median_flux_level(
+            self,
+            frame,
+            frameType="MBIAS",
+            frameName="master bias"):
+        """*calculate the median flux level in the frame, excluding masked pixels*
+
+        **Key Arguments:**
+            - ``frame`` -- the frame (CCDData object) to determine the median level.
+            - ``frameType`` -- the type of the frame for reporting QC values Default "MBIAS"
+            - ``frameName`` -- the name of the frame in human readable words. Default "master bias"
+
+        **Return:**
+            - ``medianFlux`` -- median flux level in electrons
+
+        **Usage:**
+
+        ```python
+        medianFlux = self.qc_median_flux_level(
+            frame=myFrame,
+            frameType="MBIAS",
+            frameName="master bias")
+        ```
+        """
+        self.log.debug('starting the ``qc_median_flux_level`` method')
+
+        # DETERMINE MEDIAN BIAS LEVEL
+        maskedDataArray = np.ma.array(
+            frame.data, mask=frame.mask)
+        medianFlux = np.ma.median(maskedDataArray)
+
+        utcnow = datetime.utcnow()
+        utcnow = utcnow.strftime("%Y-%m-%dT%H:%M:%S")
+
+        self.qc = self.qc.append({
+            "soxspipe_recipe": self.recipeName,
+            "qc_name": f"{frameType} MEDIAN".upper(),
+            "qc_value": medianFlux,
+            "qc_comment": f"Median flux level of {frameName}",
+            "qc_unit": "electrons",
+            "obs_date_utc": self.dateObs,
+            "reduction_date_utc": utcnow,
+            "to_header": True
+        }, ignore_index=True)
+
+        self.log.debug('completed the ``qc_median_flux_level`` method')
+        return medianFlux
+
     # use the tab-trigger below for new method
     # xt-class-method
