@@ -10,18 +10,18 @@
     March 17, 2021
 """
 ################# GLOBAL IMPORTS ####################
+from soxspipe.commonutils import keyword_lookup
+from astropy import units as u
+import ccdproc
+from astropy.nddata import CCDData
+import numpy as np
+from ._base_recipe_ import _base_recipe_
+from soxspipe.commonutils import set_of_files
+from fundamentals import tools
 from builtins import object
 import sys
 import os
 os.environ['TERM'] = 'vt100'
-from fundamentals import tools
-from soxspipe.commonutils import set_of_files
-from ._base_recipe_ import _base_recipe_
-import numpy as np
-from astropy.nddata import CCDData
-import ccdproc
-from astropy import units as u
-from soxspipe.commonutils import keyword_lookup
 
 
 class soxs_spatial_solution(_base_recipe_):
@@ -106,44 +106,45 @@ class soxs_spatial_solution(_base_recipe_):
         kw = self.kw
 
         # BASIC VERIFICATION COMMON TO ALL RECIPES
-        self._verify_input_frames_basics()
-
-        imageTypes = self.inputFrames.values(
-            keyword=kw("DPR_TYPE").lower(), unique=True)
-        imageTech = self.inputFrames.values(
-            keyword=kw("DPR_TECH").lower(), unique=True)
-        imageCat = self.inputFrames.values(
-            keyword=kw("DPR_CATG").lower(), unique=True)
+        imageTypes, imageTech, imageCat = self._verify_input_frames_basics()
 
         if self.arm == "NIR":
             # WANT ON AND OFF PINHOLE FRAMES
-            # MIXED INPUT IMAGE TYPES ARE BAD
-            if len(imageTypes) > 1:
-                imageTypes = " and ".join(imageTypes)
-                print(self.inputFrames.summary)
-                raise TypeError(
-                    "Input frames are a mix of %(imageTypes)s" % locals())
+            for i in imageTypes:
+                if i not in ["LAMP,WAVE", "LAMP,FLAT"]:
+                    raise TypeError(
+                        f"Found a {i} file. Input frames for soxspipe spatial_solution need to be LAMP,WAVE. Can optionally supply a master-flat for NIR.")
 
-            if imageTypes[0] != "LAMP,WAVE":
+            for i in imageTech:
+                if i not in ['ECHELLE,MULTI-PINHOLE', 'IMAGE', 'ECHELLE,SLIT', 'ECHELLE,PINHOLE']:
+                    raise TypeError(
+                        f"Found a {i} file. Input frames for soxspipe spatial_solution need to be LAMP,WAVE. Can optionally supply a master-flat for NIR.")
+
+            if "LAMP,WAVE" not in imageTypes:
                 raise TypeError(
                     "Input frames for soxspipe spatial_solution need to be LAMP,WAVE lamp on and lamp off frames for NIR" % locals())
 
-            for i in imageTech:
-                if i not in ['ECHELLE,MULTI-PINHOLE', 'IMAGE']:
-                    raise TypeError(
-                        "Input frames for soxspipe spatial_solution need to be LAMP,WAVE lamp on and lamp off frames for NIR" % locals())
+            if "ECHELLE,MULTI-PINHOLE" not in imageTech:
+                raise TypeError(
+                    "Input frames for soxspipe spatial_solution need to be LAMP,WAVE lamp on and lamp off frames for NIR" % locals())
 
         else:
             for i in imageTypes:
-                if i not in ["LAMP,WAVE", "BIAS", "DARK"]:
+                if i not in ["LAMP,WAVE", "BIAS", "DARK", "LAMP,FLAT"]:
                     raise TypeError(
-                        "Input frames for soxspipe spatial_solution need to be LAMP,WAVE and a master-bias and possibly a master dark for UVB/VIS" % locals())
+                        f"Found a {i} file. Input frames for soxspipe spatial_solution need to be LAMP,WAVE and a master-bias. Can optionally supply a master-flat and/or master-dark for UVB/VIS.")
 
-        # LOOK FOR ****
+        # LOOK FOR DISP MAP
         arm = self.arm
-        if arm not in self.supplementaryInput or "DISP_MAP" not in self.supplementaryInput[arm]:
+        if f"DISP_TAB_{arm}" not in imageCat:
             raise TypeError(
-                "Need a DISP_MAP for %(arm)s - none found with the input files" % locals())
+                "Need a first guess dispersion map for %(arm)s - none found with the input files" % locals())
+
+        # LOOK FOR ORDER TABLE
+        arm = self.arm
+        if f"ORDER_TAB_{arm}" not in imageCat:
+            raise TypeError(
+                "Need an order centre for %(arm)s - none found with the input files" % locals())
 
         self.imageType = imageTypes[0]
         self.log.debug('completed the ``verify_input_frames`` method')
@@ -178,7 +179,9 @@ class soxs_spatial_solution(_base_recipe_):
 
         master_bias = False
         dark = False
+        master_flat = False
         multi_pinhole_image = False
+        order_table = False
 
         add_filters = {kw("DPR_CATG"): 'MASTER_BIAS_' + arm}
         for i in self.inputFrames.files_filtered(include_path=True, **add_filters):
@@ -198,6 +201,12 @@ class soxs_spatial_solution(_base_recipe_):
             dark = CCDData.read(i, hdu=0, unit=u.adu, hdu_uncertainty='ERRS',
                                 hdu_mask='QUAL', hdu_flags='FLAGS', key_uncertainty_type='UTYPE')
 
+        # UVB/VIS/NIR FLAT
+        add_filters = {kw("DPR_CATG"): 'MASTER_LAMP-FLAT_' + arm}
+        for i in self.inputFrames.files_filtered(include_path=True, **add_filters):
+            master_flat = CCDData.read(i, hdu=0, unit=u.adu, hdu_uncertainty='ERRS',
+                                       hdu_mask='QUAL', hdu_flags='FLAGS', key_uncertainty_type='UTYPE')
+
         # MULTIPINHOLE IMAGE
         add_filters = {kw("DPR_TYPE"): 'LAMP,WAVE',
                        kw("DPR_TECH"): 'ECHELLE,MULTI-PINHOLE'}
@@ -205,8 +214,16 @@ class soxs_spatial_solution(_base_recipe_):
             multi_pinhole_image = CCDData.read(i, hdu=0, unit=u.adu, hdu_uncertainty='ERRS',
                                                hdu_mask='QUAL', hdu_flags='FLAGS', key_uncertainty_type='UTYPE')
 
-        self.multiPinholeFrame = self.subtract_calibrations(
-            inputFrame=multi_pinhole_image, master_bias=master_bias, dark=dark)
+        # FIND THE ORDER TABLE
+        filterDict = {kw("PRO_CATG").lower(): f"ORDER_TAB_{arm}"}
+        order_table = self.inputFrames.filter(**filterDict).files_filtered(include_path=True)[0]
+
+        add_filters = {kw("PRO_CATG"): f"DISP_TAB_{arm}".upper()}
+        for i in self.inputFrames.files_filtered(include_path=True, **add_filters):
+            disp_map_table = i
+
+        self.multiPinholeFrame = self.detrend(
+            inputFrame=multi_pinhole_image, master_bias=master_bias, dark=dark, master_flat=master_flat, order_table=order_table)
 
         if self.settings["save-intermediate-products"]:
             fileDir = self.intermediateRootPath
@@ -216,17 +233,51 @@ class soxs_spatial_solution(_base_recipe_):
 
         # GENERATE AN UPDATED DISPERSION MAP
         from soxspipe.commonutils import create_dispersion_map
-        mapPath = create_dispersion_map(
+        mapPath, mapImagePath = create_dispersion_map(
             log=self.log,
             settings=self.settings,
             pinholeFrame=self.multiPinholeFrame,
-            firstGuessMap=self.supplementaryInput[arm]["DISP_MAP"]
+            firstGuessMap=disp_map_table,
+            orderTable=order_table,
+            qcTable=False,
+            productsTable=False
         ).get()
 
+        from datetime import datetime
+        filename = os.path.basename(mapPath)
+
+        utcnow = datetime.utcnow()
+        utcnow = utcnow.strftime("%Y-%m-%dT%H:%M:%S")
+
+        self.products = self.products.append({
+            "soxspipe_recipe": self.recipeName,
+            "product_label": "SPAT_SOL",
+            "file_name": filename,
+            "file_type": "FITS",
+            "obs_date_utc": self.dateObs,
+            "reduction_date_utc": utcnow,
+            "product_desc": f"{self.arm} full dispersion-spatial solution",
+            "file_path": productPath
+        }, ignore_index=True)
+
+        filename = os.path.basename(mapImagePath)
+        self.products = self.products.append({
+            "soxspipe_recipe": self.recipeName,
+            "product_label": "2D_MAP",
+            "file_name": filename,
+            "file_type": "FITS",
+            "obs_date_utc": self.dateObs,
+            "reduction_date_utc": utcnow,
+            "product_desc": f"{self.arm} 2D detector map of wavelength, slit position and order",
+            "file_path": productPath
+        }, ignore_index=True)
+
+        self.report_output()
+
         self.clean_up()
 
         self.log.debug('completed the ``produce_product`` method')
-        return mapPath
+        return mapPath, mapImagePath
 
     # use the tab-trigger below for new method
     # xt-class-method
